#include <linux/limits.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <sys/wait.h>
#include <ctype.h>
#include <dirent.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <signal.h>
#include <string.h>

#ifndef PORT
#define PORT 10000
#endif

#define MAXLINE 80
#define BUFSZ	8

int logged_in = 0;
int canary_seed;

void setup_canary(char *canary)
{
	char *p;
	srand(canary_seed);
	for (p = canary; p != canary + BUFSZ; p += sizeof(int))
		*(int *)p = rand();
}

void check_canary(const char *canary)
{
	const char *p;
	srand(canary_seed);
	for (p = canary; p != canary + BUFSZ; p += sizeof(int)) {
		if (*(int *)p != rand()) {
			fprintf(stderr, "*** buffer overflow detected ***\n");
			exit(1);
		}
	}
}

void setup_permissions(int privileged, char *permissions);
int check_permission(char cmd, const char *permissions);
void do_grep(const char *pattern);
void do_ls();
// thanks to our canary implementation we can safely use gets()
char *gets(char *);
void do_exec(int privileged)
{
	char buffer[BUFSZ];
	char canary[BUFSZ];
	char permtd[BUFSZ];
	int i;
	int c;

	if (!logged_in) {
		fprintf(stderr, "you must login first\n");
		return;
	}

	// read a random value from an uninitialized local variable
	canary_seed = *(unsigned int*)canary;
	setup_canary(canary);
	setup_permissions(privileged, permtd);

	if (gets(buffer) == NULL)
		return;

	check_canary(canary);
	if (!check_permission(buffer[0], permtd)) {
		fprintf(stderr, "sorry, you need privileged access for that\n");
		return;
	}

	switch (buffer[0]) {
	case 'l':
		do_ls();
		break;
	case 'g':
		do_grep("boring");
		break;
	case 'G':
		do_grep(buffer + 1);
		break;
	default:
		fprintf(stderr, "unknown command\n");
		break;
	}
}

// NO INTENTIONAL BUGS AFTER THIS POINT
void do_login()
{
	char pw[MAXLINE];

	if (fgets(pw, MAXLINE, stdin) == NULL)
		return;
	printf("privileged access disabled, sorry\n");
	// login as unprivileged
	logged_in = 1;
}

void child(int uid)
{
	char buffer[MAXLINE];
	char *p;

	printf("Secure remote grep server v1.0\n");

	for (;;) {
		if (fgets(buffer, MAXLINE, stdin) == NULL)
			break;
		for (p = buffer; *p; p++) {
			switch (*p) {
			case '\n':
				break;
			case 'l':
				do_login();
				break;
			case 'e':
				do_exec(0);
				break;
			}
		}
	}
}

void do_ls()
{
	DIR *dir;
	struct dirent* ent;

	dir = opendir(".");

	while ( (ent = readdir(dir)) ) {
		if (ent->d_name[0] == '.')
			continue;
		printf("%s\n", ent->d_name);
	}
	closedir(dir);
}

#define MAX_FILES 20
#define FILENAMESZ 256
void do_grep(const char *pattern)
{
	char files[MAX_FILES][FILENAMESZ];
	DIR *dir;
	struct dirent* ent;
	int argc;
	char *argv[MAX_FILES + 3];

	argv[0] = "/usr/bin/grep";
	argv[1] = (char *)pattern;
	argc = 2;
	dir = opendir(".");
	while ( argc - 2 < MAX_FILES && (ent = readdir(dir)) ) {
		if (ent->d_name[0] == '.')
			continue;
		strncpy(files[argc - 2], ent->d_name, FILENAMESZ);
		argv[argc] = files[argc - 2];
		argc++;
	}
	argv[argc] = NULL;
	closedir(dir);
	if (fork()) {
		wait(0);
	} else {
		execv("/usr/bin/grep", argv);
		perror("/usr/bin/grep");
		exit(1);
	}
}

void setup_permissions(int privileged, char *permissions)
{
	char *p = permissions;
	
	*p++ = 'l'; // always allow listing
	*p++ = 'g'; // always allow grep with fixed pattern
	if (privileged) {
		// allow grep with arbitrary pattern only to administrators
		*p++ = 'G';
	}
	*p++ = '\0';
}

int check_permission(char cmd, const char *permissions)
{
	const char *p;

	for (p = permissions; *p; p++)
		if (*p == cmd)
			return 1;
	return 0;
}
void sigchld(int signo)
{
	int status;
	pid_t pid = waitpid(-1, &status, WNOHANG);
	if (pid > 0) {
		printf("child %ld: ", (long)pid);
		if (WIFEXITED(status)) {
			printf("exited with status: %d\n", WEXITSTATUS(status));
		} else {
			printf("%s\n", strsignal(WTERMSIG(status)));
		}
	}
}

int main()
{
	int lstn;
	int enable;
	struct sockaddr_in lstn_addr;

	setvbuf(stdout, NULL, _IONBF, 0);
	setvbuf(stderr, NULL, _IONBF, 0);

	lstn = socket(AF_INET, SOCK_STREAM, 0);
	if (lstn < 0) {
		perror("socket");
		return 1;
	}
	enable = 1;
	if (setsockopt(lstn, SOL_SOCKET, SO_REUSEADDR, &enable, sizeof(enable)) < 0) {
		perror("setsockopt");
		return 1;
	}
	bzero(&lstn_addr, sizeof(lstn_addr));

	lstn_addr.sin_family = AF_INET;
	lstn_addr.sin_addr.s_addr = htonl(INADDR_ANY);
	lstn_addr.sin_port = htons(PORT);

	if (bind(lstn, (struct sockaddr *)&lstn_addr, sizeof(lstn_addr)) < 0) {
		perror("bind");
		return 1;
	}

	if (listen(lstn, 10) < 0) {
		perror("listen");
		return 1;
	}
	printf("Listening on port %d\n", PORT);

	signal(SIGCHLD, sigchld);

	for (;;) {
		int con = accept(lstn, NULL, NULL);
		if (con < 0) {
			perror("accept");
			return 1;
		}

		switch (fork()) {
		case -1:
			perror("fork");
			return 1;
		case 0:
			printf("New connection, child %d\n", getpid());

			close(0);
			if (dup(con) < 0)
				goto error;
			close(1);
			if (dup(con) < 0)
				goto error;
			close(2);
			if (dup(con) < 0)
				goto error;
			child(1);
			exit(0);
			break;
		default:
			close(con);
			break;
		}
	}
	return 0;

error:
	fprintf(stderr, "dup() failed\n");
	exit(1);
}
